/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid.di

import android.util.Log
import de.chaosdorf.mete.PwaManifest
import de.chaosdorf.meteroid.model.Server
import de.chaosdorf.meteroid.model.ServerId
import de.chaosdorf.meteroid.util.await
import de.chaosdorf.meteroid.util.findBestIcon
import de.chaosdorf.meteroid.util.resolve
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.SharingStarted
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.debounce
import kotlinx.coroutines.flow.mapLatest
import kotlinx.coroutines.flow.mapNotNull
import kotlinx.coroutines.flow.stateIn
import kotlinx.serialization.ExperimentalSerializationApi
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.decodeFromStream
import okhttp3.HttpUrl.Companion.toHttpUrl
import okhttp3.OkHttpClient
import okhttp3.Request


interface AddServerViewModelFactory {
  fun newInstance(
    scope: CoroutineScope,
    isFirstServer: Boolean,
    onSubmit: (url: String, manifest: PwaManifest?) -> Unit,
    onCancel: () -> Unit
  ) = AddServerViewModelImpl(scope, isFirstServer, onSubmit, onCancel)
}

interface AddServerViewModel {
  val url: MutableStateFlow<String>
  val server: StateFlow<Server?>
  val loading: StateFlow<Boolean>
  val isFirstServer: Boolean
  fun submit()
  fun cancel()
}

class AddServerViewModelImpl(
  scope: CoroutineScope,
  override val isFirstServer: Boolean,
  private val onSubmit: (url: String, manifest: PwaManifest?) -> Unit,
  private val onCancel: () -> Unit
) : AddServerViewModel {
  private val json = Json {
    ignoreUnknownKeys = true
  }
  private val httpClient = OkHttpClient()
  override val url = MutableStateFlow("")
  private val _loading = MutableStateFlow(false)
  override val loading = _loading

  @OptIn(ExperimentalSerializationApi::class)
  private val manifest = url.debounce(300).mapNotNull { address ->
    _loading.value = true
    try {
      val url = address.toHttpUrl().resolve("manifest.json")
      val call = httpClient.newCall(Request.Builder().url(url!!).build())
      val body = call.await()
      val manifest = json.decodeFromStream<PwaManifest>(body.byteStream())
      Pair(address, manifest)
    } catch (_: Exception) {
      null
    } finally {
      _loading.value = false
    }
  }.stateIn(scope, SharingStarted.WhileSubscribed(), null)

  override val server = manifest.mapLatest { pair ->
    pair?.let { (url, manifest) ->
      Server(
        id = ServerId(-1),
        url = url,
        name = manifest.name,
        logoUrl = manifest.findBestIcon()?.resolve(url)
      )
    }
  }.stateIn(scope, SharingStarted.WhileSubscribed(), null)

  override fun submit() {
    onSubmit(url.value, manifest.value?.second)
  }

  override fun cancel() {
    onCancel()
  }
}
