/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2013-2023 Chaosdorf e.V.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package de.chaosdorf.meteroid

import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow

class RepositorySyncHandler<K, V>(
  private val withTransaction: suspend (block: suspend () -> Any?) -> Any?,
  private val repository: Repository<K, V>,
  private val loader: suspend () -> List<V>,
) : SyncHandler {
  private val _state = MutableStateFlow<SyncHandler.State>(SyncHandler.State.Idle)
  override val state: StateFlow<SyncHandler.State> = _state

  override suspend fun doSync() {
    _state.value = SyncHandler.State.Syncing
    val values =
      try {
        loader()
      } catch (e: Exception) {
        _state.value = SyncHandler.State.Error("Error while syncing: $e")
        return
      }
    withTransaction {
      val existing = repository.getAll().map(repository::getKey).toSet()
      val deletedEntries = existing - values.map(repository::getKey).toSet()
      for (deletedEntry in deletedEntries) {
        repository.delete(deletedEntry)
      }
      for (entry in values) {
        repository.save(entry)
      }
    }
    _state.value = SyncHandler.State.Idle
  }
}
